@@ -6,6 +6,7 @@ from PIL import Image
 from scipy import misc
 from .kafka_helper import KAFKA_SEARCH_TOPIC
 import os
+from scipy.interpolate import PchipInterpolator
 
 # system default image type
 DEFAULT_IMAGE_TYPE = 'jpeg'
@@ -24,12 +25,24 @@ CROPPED_TARGET_PATH = '/images/target_cropped'
 # path to save cropped faces from video process (matched and unmatched):
 MATCHED_VIDEO_PROCESSED_FACE_PATH = '/images/video_processed/matched'
 UNMATCHED_VIDEO_PROCESSED_FACE_PATH = '/images/video_processed/unmatched'
+# constants for google face model
+GFACE_A = -17.30
+GFACE_B = 4.62
 
 
 class CypressCommonHelper:
+
     def __init__(self):
         pass
 
+    @staticmethod
+    def init_pchi():
+        # used by sigmoid function
+        x = np.array([0.001, 0.200, 0.3275, 0.3862, 0.4200, 0.4658, 0.5500, 0.6500, 0.7, 1.0, 1.5, 1.80, 2.00])
+        y = np.array([100.0, 99.0, 95.0000, 90.000, 88.000, 58.000, 55.000, 45.000, 0.3, 0.2, 0.1, 0.05, 0.01])
+        pchi = PchipInterpolator(x, y)
+        return pchi
+
     @staticmethod
     def np_bytes_to_np_array(np_bytes, axis_0, axis_1):
         """
@@ -60,21 +73,31 @@ class CypressCommonHelper:
             return "cannot flat numpy array to numpy array bytes"
 
     @staticmethod
-    def img_rgb2bgr(image_rgb):
+    def img_rgb2bgr(im):
         """
-        Convert a 'RGB' image to a 'BGR' image.
-        :param image_rgb: the image object that is in 'RGB' or 'RGBA' mode
-        :type image_rgb: `~PIL.Image.Image` object
+        Convert an image to a 'BGR' image.
+        :param im: the image object that is not in 'BGR' mode
+        :type im: `~PIL.Image.Image` object
         :return: the image object in 'BGR' mode
         :rtype: `~PIL.Image.Image` object
         """
-        if image_rgb.mode == 'RGBA':
-            r, g, b, _ = image_rgb.split()
-        elif image_rgb.mode == 'RGB':
-            r, g, b = image_rgb.split()
-        else:
-            raise IOError('Input image is either RGBA or RGB image.')
+        if not Image.isImageType(im):
+            raise TypeError("Input is not a PIL image.")
 
+        if im.mode == '1':
+            # Workaround for crash in PIL. When im is 1-bit, the call array(im)
+            # can cause a seg. fault, or generate garbage. See
+            # https://github.com/scipy/scipy/issues/2138 and
+            # https://github.com/python-pillow/Pillow/issues/350.
+            #
+            # This converts im from a 1-bit image to an 8-bit image.
+            im = im.convert('L')
+
+        if 'RGB' != im.mode:
+            # Always convert the input image to mode 'RGB'
+            im = im.convert('RGB')
+
+        r, g, b = im.split()
         image_bgr = Image.merge("RGB", (b, g, r))
         return image_bgr
 
@@ -93,22 +116,15 @@ class CypressCommonHelper:
         return misc.fromimage(image_rgb)
 
     @staticmethod
-    def img_rgb_2_np_bgr(image_rgb):
+    def img_rgb_2_np_bgr(im):
         """
-        Convert a 'RGB' image object to a 'BGR' image numpy array.
-        :param image_rgb: the image object that is in 'RGB' or 'RGBA' mode
-        :type image_rgb: `~PIL.Image.Image` object
+        Convert an image object to a 'BGR' image numpy array.
+        :param im: the image object that is not in 'BGR' mode
+        :type im: `~PIL.Image.Image` object
         :return: numpy array of the image in mode 'BGR'
         :rtype: numpy array
         """
-        if image_rgb.mode == 'RGBA':
-            r, g, b, _ = image_rgb.split()
-        elif image_rgb.mode == 'RGB':
-            r, g, b = image_rgb.split()
-        else:
-            raise IOError('Input image is either RGBA or RGB image.')
-
-        image_bgr = Image.merge("RGB", (b, g, r))
+        image_bgr = CypressCommonHelper.img_rgb2bgr(im)
         np_bgr = misc.fromimage(image_bgr)
         return np_bgr
 
@@ -180,21 +196,24 @@ class CypressCommonHelper:
         os.remove(file_name)
 
     @staticmethod
-    def reload_search_engine(kafka_producer_obj, cache, timeout, sleep_time, task_name=None):
+    def reload_search_engine(kafka_producer_obj, cache, timeout, sleep_time, category_ids=None):
         """
         Reload the in memory profiles database for search engine, to get the latest profiles data from Redis database.
         :param kafka_producer_obj: PythonKafkaProducer object from kafka_helper
         :param cache: CypressCache object
         :param timeout: server's timeout in seconds, the time to wait for the result of reload request
         :param sleep_time: the time interval for each result check, in seconds
-        :param task_name: 'create_profile', 'delete_profile', 'db_import', etc. Default to None.
+        :param category_ids: a list of category ids (each category is one search instance) to reload.
+        If input is not a list, then convert it into a list of one element.
         :return: (True, None) if reload successfully; (False, '<failure reason>') if reload failed.
         :rtype: tuple
         """
         task_id = str(uuid.uuid4())
         start_time = time.time()
         try:
-            msg = {'usage': "reload", 'task_id': task_id, 'task_name': task_name}
+            if category_ids and not isinstance(category_ids, list):
+                category_ids = [category_ids]
+            msg = {'usage': "reload", 'task_id': task_id, 'category_ids': category_ids}
 
             kafka_producer_obj.send_message(KAFKA_SEARCH_TOPIC, msg)
             while time.time() - start_time < timeout:
@@ -218,3 +237,23 @@ class CypressCommonHelper:
     @staticmethod
     def get_current_time_ms():
         return int(round(time.time() * 1000))
+
+    @staticmethod
+    def compute_distance_and_sim(pchi, feat1, feat2, use_gface=False):
+        """
+        Compute distance and similarity between two feature vectors
+        :param pchi: PCHIP 1-d monotonic cubic interpolation. Get the pchi by calling CypressCommonHelper.init_pchi()
+        :param feat1: a float array of length 512
+        :param feat2: a float array of length 512
+        :param use_gface: True if uses google face model; False if uses cmu model
+        :return: distance and similarity between two vectors. The greater the distance, the lower the similarity
+        :rtype: float
+        """
+        distance = 1 - np.dot(feat1, feat2) / np.linalg.norm(feat1) / np.linalg.norm(feat2)
+        if use_gface is True:
+            A = GFACE_A
+            B = GFACE_B
+            sim = 100.0 / (1 + np.exp(-A * distance - B))
+        else:
+            sim = pchi(distance)
+        return distance, sim
