@@ -31,6 +31,8 @@ DB_IMPORT_SUCCESS = 'success'
 DB_IMPORT_TOTAL = 'total'
 DB_IMPORT_TEMP_FAIL_PREFIX = 'temp_failed_import'
 
+MIGRATION_PROFILE_HASH_NAME = 'migration:profile'
+MIGRATION_TARGET_HASH_NAME = 'migration:target'
 
 
 DATA_IMPORT_FAIL_TYPES = {
@@ -213,6 +215,37 @@ class CypressCache(CypressBase):
                 redis_retry_counter += 1
                 self.connect_to_redis()
 
+    def get_sorted_set_rank(self, name, value):
+        """
+        Returns a 0-based value indicating the rank of ``value`` in sorted set ``name``
+        :param name: sorted set name
+        :param value: the values to get rank
+        :return: If member exists in the sorted set, return the rank of member (integer).
+                If member does not exist in the sorted set or key does not exist, return None
+        """
+        redis_retry_counter = 0
+        while True:
+            try:
+                return self.redis.zrank(name, value)
+            except redis.ConnectionError:
+                redis_retry_counter += 1
+                self.connect_to_redis()
+
+    def remove_from_sorted_set(self, name, *values):
+        """
+        "Remove member ``values`` from sorted set ``name``"
+        :param name: sorted set name
+        :param values: the values to be removed
+        :return: The number of members removed from the sorted set, not including non existing members.
+        """
+        redis_retry_counter = 0
+        while True:
+            try:
+                return self.redis.zrem(name, *values)
+            except redis.ConnectionError:
+                redis_retry_counter += 1
+                self.connect_to_redis()
+
     def set_mkv_to_hash(self, name, mapping):
         """
         Set key to value within hash ``name`` for each corresponding key and value from the ``mapping`` dict.
@@ -263,7 +296,7 @@ class CypressCache(CypressBase):
         :param name: the sorted set name
         :type name: basestring
         :param values: the values to delete
-        :type values: string separated by comma
+        :type values: string separated by comma. i.e. r.zrem('test_set', 'b:1', 'b:4')
         :return: The number of members removed from the sorted set, not including non existing members.
         :rtype: int
         """
@@ -310,28 +343,6 @@ class CypressCache(CypressBase):
     ##############################################################
     ###### Video Processor Functions ######
     ##############################################################
-    def add_task_to_worker(self, worker_id, task_id, task):
-        """
-        Add an offline processor task
-        :param worker_id: the id of the offline processor worker
-        :type worker_id: basestring or uuid
-        :param task: video file name with extension
-        :type task: basestring
-        :return:
-        """
-        key = PROCESSOR_TASK_PREFIX + str(worker_id)
-        self.set_kv_to_hash(key, "task_id", task_id)
-        self.set_kv_to_hash(key, "task", task)
-
-    def get_task_by_worker_id(self, worker_id):
-        """
-        Get the task by worker id
-        :param worker_id: the id of the offline processor worker
-        :type worker_id: basestring or uuid
-        :return:
-        """
-        key = PROCESSOR_TASK_PREFIX + str(worker_id)
-        return self.get_kv_by_hash_name(key, "task")
 
     def get_task_id_by_worker_id(self, worker_id):
         """
@@ -343,6 +354,44 @@ class CypressCache(CypressBase):
         key = PROCESSOR_TASK_PREFIX + str(worker_id)
         return self.get_kv_by_hash_name(key, "task_id")
 
+    # The below methods are for VA-3147
+    def add_task_to_worker(self, worker_id, task_id, task, video_id=None, category_ids=None):
+        """
+        Add an offline processor task
+        :param worker_id: the id of the offline processor worker
+        :type worker_id: basestring or uuid
+        :param task: video file name with extension
+        :type task: basestring
+        :return:
+        """
+        key = PROCESSOR_TASK_PREFIX + str(worker_id)
+        self.set_kv_to_hash(key, "task_id", task_id)
+        self.set_kv_to_hash(key, "task", task)
+        self.set_kv_to_hash(key, "video_id", video_id)
+        self.set_kv_to_hash(key, "category_ids", category_ids)
+
+    def get_task_by_worker_id(self, worker_id):
+        """
+        Get the task by worker id
+        :param worker_id: the id of the offline processor worker
+        :type worker_id: basestring or uuid
+        :return:
+        """
+        key = PROCESSOR_TASK_PREFIX + str(worker_id)
+        task = None
+        task_id = None
+        video_id = None
+        category_ids = None
+
+        if self.is_key_exists(key):
+            task_hash_map = self.get_all_by_hash_name(key)
+            task = task_hash_map.get("task")
+            task_id = task_hash_map.get("task_id")
+            video_id = task_hash_map.get("video_id")
+            category_ids = task_hash_map.get("category_ids")
+
+        return task, task_id, video_id, category_ids
+
     def remove_worker_task(self, worker_id):
         """
         Delete the worker task
@@ -353,6 +402,18 @@ class CypressCache(CypressBase):
         key = PROCESSOR_TASK_PREFIX + str(worker_id)
         self.delete_keys(key)
 
+    def get_all_worker_ids(self):
+        """
+        Get worker id for all video processing task.
+        :return: A list of worker id for video processing tasks.
+        """
+        worker_ids = []
+        for keys in self.redis.scan_iter(match=PROCESSOR_TASK_PREFIX + '*'):
+            worker_id = keys.split(":")
+            if len(worker_id) == 2:
+                worker_ids.append(worker_id[-1])
+        return worker_ids
+
     def set_task_total_frames(self, task_id, total):
         key = VIDEO_TOTAL_FRAMES_PREFIX+str(task_id)
         self.set_value_to_key(key, total)
@@ -525,7 +586,8 @@ class CypressCache(CypressBase):
         -1. delete from profile vector hash table
         -2. delete from profile category hash table
         -3. delete profile np hash table
-        -4. delete profile metadata from profile metadata set
+        -4. delete profile metadata from profile metadata key value pair
+        -5. delete profile metadata from profile metadata set
         :param category_id: The id of the category to delete
         :type category_id: basestring
         :return: True if cleared successfully (no matter it exist or not)
@@ -535,17 +597,14 @@ class CypressCache(CypressBase):
             self.logger.warning('Category does not exist in category table: {0}'.format(category_id))
 
         profile_ids = self.get_profile_ids_by_category(category_id)
-
         if profile_ids:
             cat_profile_ids = []
             np_profile_hashes = []
-            set_name = METADATA_PREFIX + category_id
+            meta_keys = []
             for image_id in profile_ids:
                 cat_profile_ids.append(category_id + ':' + image_id)
                 np_profile_hashes.append(PROFILE_NP_PREFIX + image_id)
-                for value, score in self.redis.zscan_iter(set_name, match=image_id + ":*"):
-                    # 4. delete profile metadata from profile metadata set
-                    self.delete_from_sorted_set(set_name, value)
+                meta_keys.append(METADATA_PREFIX + image_id)
 
             # 1. delete from profile vector hash table
             self.delete_keys_from_hash(VECTOR_HASH_NAME, *cat_profile_ids)
@@ -553,6 +612,11 @@ class CypressCache(CypressBase):
             self.delete_keys_from_hash(CATEGORY_HASH_NAME, *profile_ids)
             # 3. delete from profile np hash table
             self.delete_keys(*np_profile_hashes)
+            # 4. delete profile metadata from profile metadata key value pair
+            self.delete_keys(*meta_keys)
+            # 5. delete profile metadata from profile metadata set
+            set_name = METADATA_PREFIX + category_id
+            self.delete_keys(set_name)
         else:
             self.logger.info('No profile is found for category: {0}'.format(category_id))
 
@@ -564,8 +628,9 @@ class CypressCache(CypressBase):
         -1. delete from profile vector hash table
         -2. delete from profile category hash table
         -3. delete profile np hash table
-        -4. delete profile metadata from profile metadata set
-        -5. delete profile category table
+        -4. delete profile metadata from profile metadata key value pair
+        -5. delete profile metadata from profile metadata set
+        -6. delete profile category table
         :param category_id: The id of the category to delete
         :type category_id: basestring
         :return: True if deleted successfully (no matter it exist or not), False if delete failed
@@ -575,17 +640,14 @@ class CypressCache(CypressBase):
             self.logger.warning('Category does not exist in category table: {0}'.format(category_id))
 
         profile_ids = self.get_profile_ids_by_category(category_id)
-
         if profile_ids:
             cat_profile_ids = []
             np_profile_hashes = []
-            set_name = METADATA_PREFIX + category_id
+            meta_keys = []
             for image_id in profile_ids:
                 cat_profile_ids.append(category_id + ':' + image_id)
                 np_profile_hashes.append(PROFILE_NP_PREFIX + image_id)
-                for value, score in self.redis.zscan_iter(set_name, match=image_id+":*"):
-                    # 4. delete profile metadata from profile metadata set
-                    self.delete_from_sorted_set(set_name, value)
+                meta_keys.append(METADATA_PREFIX + image_id)
 
             # 1. delete from profile vector hash table
             self.delete_keys_from_hash(VECTOR_HASH_NAME, *cat_profile_ids)
@@ -593,10 +655,15 @@ class CypressCache(CypressBase):
             self.delete_keys_from_hash(CATEGORY_HASH_NAME, *profile_ids)
             # 3. delete from profile np hash table
             self.delete_keys(*np_profile_hashes)
+            # 4. delete profile metadata from profile metadata key value pair
+            self.delete_keys(*meta_keys)
+            # 5. delete profile metadata from profile metadata set
+            set_name = METADATA_PREFIX + category_id
+            self.delete_keys(set_name)
         else:
             self.logger.info('No profile is found for category: {0}'.format(category_id))
 
-        # 5. delete from profile category table
+        # 6. delete from profile category table
         cat_key = CATEGORY_PREFIX + category_id
         delete_task = self.delete_keys(cat_key)
         return delete_task == 1 or delete_task == 0
@@ -610,14 +677,12 @@ class CypressCache(CypressBase):
         :return: the number of profiles within the category
         """
         if category_id is None:
-            return len(self.redis.hkeys(VECTOR_HASH_NAME))
+            return self.redis.hlen(VECTOR_HASH_NAME)
 
         category_id = str(category_id)
-        count = 0
-        for k in self.redis.hkeys(VECTOR_HASH_NAME):
-            cat_id = k.split(':')[0]
-            if category_id == cat_id:
-                count += 1
+        set_name = METADATA_PREFIX + category_id
+        count = self.redis.zcard(set_name)
+
         return count
 
     def add_image_to_profiles(self, image_id, category_id, np_bytes, axis0, axis1, expire=600):
@@ -721,12 +786,15 @@ class CypressCache(CypressBase):
 
     def add_metadata_to_profile(self, profile_id, category_id, metadata, max_size=128):
         """
-        Add user metadata to a profile, as sorted set for pagination purpose. Only can create a new one, not update!
+        Add user metadata to a profile, as sorted set for pagination purpose.
+        If the profile with same metadata already exists, the new request will overwrite the old data.
         Only call this method after add_image_to_profiles(profile_id, ...) was called, make sure the profile id exists in:
         { "profile:category": {"<profile_id3>": "<category_id2>", ......} }
 
         i.e. metadata structure
-        { "profile:metadata:<category_id1>": ["<image_id1>:user_metadata_1"(score1), "<image_id3>:user_metadata_2"(score2), ......] }
+        { "profile:metadata:<category_id1>": ["<image_id1>"(score1), "<image_id3>"(score2), ......] }
+        For each profile that has metadata:
+        { "profile:metadata:<image_id1>": "user_metadata_1"}
         :param profile_id: the id of the profile/image to add user metadata to.
         :type profile_id: basestring or UUID
         :param category_id: the id of the category the profile belongs to.
@@ -735,7 +803,7 @@ class CypressCache(CypressBase):
         :type metadata: basestring
         :param max_size: maximum size of input metadata, default to 128
         :type max_size: int
-        :return: True if add metadata to profile succeed; False if metadata already exists for this profile.
+        :return: True if add/update metadata to profile succeed; False otherwise.
         """
         if isinstance(metadata, unicode):
             metadata = metadata.encode('utf-8')
@@ -749,16 +817,14 @@ class CypressCache(CypressBase):
             raise IOError('The profile: {} does not belong to category: {}.'.format(profile_id, category_id))
 
         set_name = METADATA_PREFIX + category_id
-        for value, score in self.redis.zscan_iter(set_name, match=profile_id+':*'):
-            if value:
-                self.logger.error('Profile: {} already have metadata.'.format(profile_id))
-                return False
-
-        set_value = profile_id + ':' + metadata
-        # generate a random float number between 0 and 1000
+        # generate a random float number between 0 and 1000. If it's updating, the score/rank of the value will change
         random_score = random.uniform(0, 1000)
-        result = self.add_to_sorted_set(set_name, set_value, random_score)
-        return result == 1
+        self.add_to_sorted_set(set_name, profile_id, random_score)
+
+        metadata_key = METADATA_PREFIX + profile_id
+        result = self.set_value_to_key(metadata_key, str(metadata))
+
+        return result == 1 or result == 0
 
     def get_profiles_by_category(self, category_id, start, end, desc=False, withscores=False, score_cast_func=float):
         """
@@ -769,8 +835,8 @@ class CypressCache(CypressBase):
         :param desc: a boolean indicating whether to sort the results descendingly
         :param withscores: indicates to return the scores along with the values. The return type is a list of (value, score) pairs
         :param score_cast_func: a callable used to cast the score return value
-        :return: Empty list if there's no profiles within the category; Otherwise, a list of metadata with image id as prefix:
-        i.e. ["<image_id1>:user_metadata_1", "<image_id3>:user_metadata_2", ......]
+        :return: Empty list if there's no profiles within the category; Otherwise, a list of metadata with image id:
+        i.e. ["<image_id1>", "<image_id3>", ......]
         """
         category_id = str(category_id)
         set_name = METADATA_PREFIX + category_id
@@ -779,23 +845,14 @@ class CypressCache(CypressBase):
 
     def get_profile_metadata(self, profile_id):
         """
-        Get user metadata of a profile.
+        Get user metadata of a profile. i.e.
+        { "profile:metadata:<image_id1>": "user_metadata_1"}
         :param profile_id: the id of the profile to get metadata from
         :type profile_id: basestring or uuid
         :return: profile user metadata as a string; None if no metadata is found.
         :rtype: basestring
         """
-        profile_id = str(profile_id)
-        category_id = self.get_category_id(profile_id)
-        if not category_id:
-            self.logger.error('No category is found for profile: {}'.format(profile_id))
-            return None
-        set_name = METADATA_PREFIX + category_id
-        for value, score in self.redis.zscan_iter(set_name, match=profile_id + ':*'):
-            if value:
-                # remove the image id part from the value
-                return ':'.join(value.split(':')[1:])
-        return None
+        return self.get_value_by_key(METADATA_PREFIX + str(profile_id))
 
     def add_vector_to_profile(self, image_id, vector):
         """
@@ -915,11 +972,12 @@ class CypressCache(CypressBase):
     def delete_profile(self, image_id):
         """
         Remove an image (np data, vector data, categroy data) in profiles database.
-        The four data structures to operate on:
+        The five data structures to operate on:
         { "profile:vector": {"<category_id1>:<image_id1>": "vector2", "<category_id2>:<image_id3>": "vector3", ......} }
         { "profile:np:<image_id1>": {"matrix":"<numpy array bytes>", "axis0": "2", "axis1": "5"} }
         { "profile:category": {"<image_id1>": "<category id>", "<image_id3>": "<category id>", ......} }
-        { "profile:metadata:<category_id1>": ["<image_id1>:user_metadata_1"(score1), "<image_id3>:user_metadata_2"(score2), ......] }
+        { "profile:metadata:<image_id1>": "user_metadata_1" }
+        { "profile:metadata:<category_id1>": ["<image_id1>"(score1), "<image_id3>"(score2), ......] }
         :param image_id: the id of the image to delete
         :type image_id: basestring or uuid
         :return: (True, 204) if deleted; (False, 404) if image does not exist in profiles database
@@ -936,9 +994,10 @@ class CypressCache(CypressBase):
         del_vector = self.delete_keys_from_hash(VECTOR_HASH_NAME, vector_key)
         del_cat = self.delete_keys_from_hash(CATEGORY_HASH_NAME, image_id)
 
+        metadata_key = METADATA_PREFIX + image_id
+        self.delete_keys(metadata_key)
         set_name = METADATA_PREFIX + category_id
-        for value, score in self.redis.zscan_iter(set_name, match=image_id+':*'):
-            self.delete_from_sorted_set(set_name, value)
+        self.delete_from_sorted_set(set_name, image_id)
 
         # note: this returns either 0 or 1 because np_hash_name could be expired
         del_np = self.delete_keys(np_hash_name)
@@ -1121,7 +1180,7 @@ class CypressCache(CypressBase):
         hash_name = ENGINE_SEARCH_HASH_NAME_PREFIX + task_id
         obj = self.get_all_by_hash_name(hash_name)
         if not obj:
-            self.logger.debug('Engine search task information does not exist: {0}'.format(task_id))
+            # self.logger.debug('Engine search task information does not exist: {0}'.format(task_id))
             return None
         return obj
 
@@ -1169,7 +1228,7 @@ class CypressCache(CypressBase):
         hash_name = ENGINE_DETECTION_HASH_NAME_PREFIX + task_id
         obj = self.get_all_by_hash_name(hash_name)
         if not obj:
-            self.logger.debug('Engine detection task information does not exist: {0}'.format(task_id))
+            # self.logger.debug('Engine detection task information does not exist: {0}'.format(task_id))
             return None
         return obj
 
@@ -1225,7 +1284,7 @@ class CypressCache(CypressBase):
         hash_name = ENGINE_RECOGNITION_HASH_NAME_PREFIX + task_id
         obj = self.get_all_by_hash_name(hash_name)
         if not obj:
-            self.logger.debug('Engine recognition task information does not exist: {0}'.format(task_id))
+            # self.logger.debug('Engine recognition task information does not exist: {0}'.format(task_id))
             return None
         return obj
 
@@ -1413,4 +1472,70 @@ class CypressCache(CypressBase):
         get DB_IMPORT_TEMP_FAIL_PREFIX count for unstructured import
         :return: count
         """
-        return self.get_value_by_key(DB_IMPORT_TEMP_FAIL_PREFIX)
\ No newline at end of file
+        return self.get_value_by_key(DB_IMPORT_TEMP_FAIL_PREFIX)
+
+    def set_migration_stats(self, status, count, usage='profile'):
+        """
+        Update the processed stats for data migration service, including the counts for success, fail, and total number of records.
+        :param status: the key in the hash mapping: 'success', 'error', 'total'
+        :type status: basestring
+        :param count: the number of records being processed for the specified status
+        :type count: int
+        :param usage: either profile or target
+        :type usage: basestring
+        :return: True if set successfully, False if set failed.
+        """
+        status_list = [DB_IMPORT_ERROR, DB_IMPORT_SUCCESS, DB_IMPORT_TOTAL]
+        if status not in status_list:
+            self.logger.error('Failed to set data migration stats, the input status does not exist: {0}'.format(status))
+            raise IOError('Unrecognized input status value.')
+        count = int(count)
+        # create or update the count for status
+        if usage == 'profile':
+            set_count = self.set_kv_to_hash(MIGRATION_PROFILE_HASH_NAME, status, count)
+        else:
+            set_count = self.set_kv_to_hash(MIGRATION_TARGET_HASH_NAME, status, count)
+        return set_count == 1 or set_count == 0
+
+    def increment_migration_stats(self, status, amount, usage='profile'):
+        """
+        Increment the processed stats for data migration service, including the counts for success and fail number of records.
+        If key(status) does not exist the value is set to 0 before the operation is performed.
+        :param status: the key in the hash mapping: 'success', 'error'
+        :type status: basestring
+        :param amount: the number to increment for the value of the key
+        :type amount: int
+        :param usage: either profile or target
+        :type usage: basestring
+        :return: The new value of the key
+        """
+        status_list = [DB_IMPORT_ERROR, DB_IMPORT_SUCCESS]
+        if status not in status_list:
+            self.logger.error('Failed to set data migration stats, the input status does not exist: {0}'.format(status))
+            raise IOError('Unrecognized input status value.')
+        amount = int(amount)
+        if usage == 'profile':
+            return self.increment_hash_kv(MIGRATION_PROFILE_HASH_NAME, status, amount)
+        else:
+            return self.increment_hash_kv(MIGRATION_TARGET_HASH_NAME, status, amount)
+
+    def get_migration_stats(self, usage='profile'):
+        """
+        Get the stats for data migration service.
+        :param status Default to None.
+        :type status basestring
+        :param usage: either profile or target
+        :type usage: basestring
+        :return: a dictionary contains all status.
+        i.e.  {"success": "0", "error": "0", "total": "0"}
+        None if no data.
+        """
+        if usage == 'profile':
+            obj = self.get_all_by_hash_name(MIGRATION_PROFILE_HASH_NAME)
+        else:
+            obj = self.get_all_by_hash_name(MIGRATION_TARGET_HASH_NAME)
+
+        if not obj:
+            self.logger.warning('Data import stats does not exist.')
+            return None
+        return obj
