@@ -1,20 +1,22 @@
 import Queue
 import json
-import threading
 import traceback
-from thread import start_new_thread
 from threading import Thread
 from time import sleep, time
 import confluent_kafka
 from kafka.errors import KafkaTimeoutError
-
 from .cypress_base import CypressBase
 from kafka import KafkaProducer
 
-# search engine consumes messages with topic "search" and produces messages with topic "search_mw"
+
+# search engine manager consumes messages with topic "search" and produces messages with topic "search_mw"
 KAFKA_SEARCH_TOPIC = "search"
 KAFKA_SEARCH_RESULT_TOPIC = "search_mw"
 
+# search engine consumes messages with topic "sub_search" and produces messages with topic "sub_search_mw"
+KAFKA_SUB_SEARCH_TOPIC = "sub_search"
+KAFKA_SUB_SEARCH_RESULT_TOPIC = "sub_search_mw"
+
 # engine face detection service consumes messages with topic "face_detection"
 #  and produces messages with topic "face_detection_mw"
 KAFKA_DETECTION_TOPIC = "face_detection"
@@ -121,7 +123,7 @@ class PythonKafkaProducer(CypressBase):
 
 
 class KafkaProducerBase(Thread, CypressBase):
-    def __init__(self, bootstrap_server, config=None, message_queue=None):
+    def __init__(self, bootstrap_server, config=None):
         """
         :param bootstrap_server:
         :param config:
@@ -131,21 +133,25 @@ class KafkaProducerBase(Thread, CypressBase):
         Thread.__init__(self)
         self.bootstrap_server = bootstrap_server
         self.running = True
+        self.message_queue = Queue.Queue()
         self.config = config if config else {}
         self.config["bootstrap.servers"] = self.bootstrap_server
         self.config["socket.keepalive.enable"] = True
         self.config["topic.metadata.refresh.interval.ms"] = 10000
-        self.config["queue.buffering.max.ms"] = 10
         self.config["message.send.max.retries"] = 10
+        if "socket.blocking.max.ms" not in self.config.keys():
+            self.config["socket.blocking.max.ms"] = 1
+        if "queue.buffering.max.ms" not in self.config.keys():
+            self.config["queue.buffering.max.ms"] = 1
+        self.debug_mode = self.get_debug_mode()
         self.producer = self.try_connect_kafka()
-        self.message_queue = message_queue if message_queue else Queue.Queue()
+        self.logger.debug("producer configs: {}".format(self.config))
         self.logger.debug("init KafkaProducerBase spent {0}".format(time() - t1))
 
     def run(self):
-        # start_new_thread(self.producer_poll, ())
         while self.running:
             try:
-                topic, msg = self.message_queue.get(block=False, timeout=1000)
+                topic, msg = self.message_queue.get(block=False, timeout=10)
             except:
                 continue
             if self.producer is None:
@@ -156,21 +162,36 @@ class KafkaProducerBase(Thread, CypressBase):
                         "Cannot send message to {0}:{1}".format(self.bootstrap_server, topic))
                     return
                 self.logger.debug("Create kafka producer spends {0}s".format(time() - start_time))
-            start = time()
             self.produce_message(topic, msg)
-            self.logger.debug("sent messages to topic [{0}] spent {1}s".format(topic, time() - start))
+
+    def send_message(self, topic, msg=None, message=None):
+        if msg:
+            self.message_queue.put((topic, msg))
+        elif message:
+            self.message_queue.put((topic, message))
+        else:
+            raise Exception("message/msg cannot be empty!")
 
     def produce_message(self, topic, msg):
         retry = 10
+        if self.producer is None:
+            start_time = time()
+            self.producer = self.try_connect_kafka()
+            if self.producer is None:
+                self.logger.error(
+                    "Cannot send message to {0}:{1}".format(self.bootstrap_server, topic))
+                return
+            self.logger.debug("Create kafka producer spends {0}s".format(time() - start_time))
         while retry:
             try:
+                msg["start_ts"] = time()
                 self.producer.produce(topic, value=json.dumps(msg))
-                self.producer.poll(0)
+                self.producer.poll(1)
                 return
             except BufferError as e:
                 self.logger.error(e)
                 retry -= 1
-                self.producer.poll(0)
+                self.producer.poll(1)
             except Exception as ex:
                 self.logger.error(traceback.format_exc())
                 self.logger.error(ex)
@@ -180,7 +201,6 @@ class KafkaProducerBase(Thread, CypressBase):
                 return
             except KeyboardInterrupt:
                 self.logger.info("Received exit input. Exiting...")
-                self.running = False
         self.logger.error("Failed to send message to {0}:{1}".format(self.bootstrap_server, topic))
         self.producer = None
 
@@ -218,8 +238,8 @@ class KafkaConsumerBase(Thread, CypressBase):
     each subclass need to implement its own message_handler
     .. note:: this class has to be extended to some real functionality.
     """
-    def __init__(self, bootstrap_server, topic, partition=0, group="group1", polling_timeout=0.1, connect_retries=100,
-                 reconnect_interval=0.5, config=None):
+    def __init__(self, bootstrap_server, topic, partition=0, group="cypress_default", polling_timeout=30,
+                 connect_retries=100, reconnect_interval=0.5, config=None):
         """
         :param bootstrap_server: server ip:port
         :param topic: consumer topic
@@ -237,7 +257,6 @@ class KafkaConsumerBase(Thread, CypressBase):
         self.config = config if config else {}
         self.config["bootstrap.servers"] = bootstrap_server
         self.config["topic.metadata.refresh.interval.ms"] = 10000
-        self.config["queued.min.messages"] = 10
         self.config["group.id"] = group
         self.consumer_topic = topic
         self.producer_topic = topic+"_mw"
@@ -245,19 +264,19 @@ class KafkaConsumerBase(Thread, CypressBase):
         self.polling_timeout = polling_timeout
         self.reconnect_interval = reconnect_interval
         self.connect_retries = connect_retries
-        self.message_queue = Queue.Queue()
-        self.kafka_producer = None
         self.kafka_consumer = None
+
+        if "fetch.wait.max.ms" not in self.config.keys():
+            self.config["fetch.wait.max.ms"] = 0
+        if "socket.blocking.max.ms" not in self.config.keys():
+            self.config["socket.blocking.max.ms"] = 1
         self.try_connect_kafka()
+        self.reconnect = False
+        self.logger.debug("consumer configs: {}".format(self.config))
         self.logger.debug("init KafkaConsumerBase spent {0} sec".format(time() - t1))
 
     def stop(self):
         self.running = False
-        if self.kafka_producer:
-            try:
-                self.kafka_producer.stop()
-            except:
-                pass
 
     def __exit__(self, exc_type, exc_val, exc_tb):
         """
@@ -281,44 +300,63 @@ class KafkaConsumerBase(Thread, CypressBase):
 
         """
 
-        self.kafka_consumer = None
         while True:
-            if self.kafka_consumer:
-                break
             try:
+
                 self.kafka_consumer = confluent_kafka.Consumer(**self.config)
                 # use partition for consume messages. The partition number is determined by the client io
                 # self.kafka_consumer.assign([confluent_kafka.cimpl.TopicPartition(self.consumer_topic, self.partition)])
                 self.kafka_consumer.subscribe([self.consumer_topic])
+                self.logger.debug("connected to Kafka server, testing topic connection...")
+                if not self.test_kafka_connection():
+                    raise Exception("cannot connect to topic, reconnecting...")
+                else:
+                    self.reconnect = False
+                    self.logger.info("~~~~~~~~kafka consumer for topic {} connected~~~~~~~~".format(self.consumer_topic))
+                    break
 
-            # TODO: catch different types of exception, dealing with different strategies
             except:
                 self.logger.error(traceback.format_exc())
                 # wait 0.1 second to retry
                 sleep(self.reconnect_interval)
-                self.kafka_consumer = None
+                try:
+                    self.kafka_consumer.close()
+                except:
+                    pass
+
+    def test_kafka_connection(self):
+        retry = 30
+        connected = False
+        while retry:
+            message = self.kafka_consumer.poll(1)
+            if message is None:
+                retry -= 1
+                continue
+            if message.error():
+                # Error or event
+                if message.error().code() == confluent_kafka.KafkaError._PARTITION_EOF:
+                    connected = True
+                    break
+
+        return connected
 
     def run(self):
-
         """
         main loop.
 
         """
-
         while self.running:
             try:
-                if self.kafka_consumer is None:
+                if self.kafka_consumer is None or self.reconnect:
                     self.try_connect_kafka()
-                    if self.kafka_consumer is None:
-                        self.logger.error("Run: cannot connect to Kafka server.")
-                        continue
-                    else:
-                        self.logger.info("Kafka Consumer for topic: {0} connected".format(self.consumer_topic))
-                message = self.kafka_consumer.poll(timeout=self.polling_timeout)
+
+                # polling timeout is 30 seconds
+                message = self.kafka_consumer.poll(self.polling_timeout)
                 if not self.running:
                     break
                 if message is None:
                     continue
+
                 if message.error():
                     # Error or event
                     if message.error().code() == confluent_kafka.KafkaError._PARTITION_EOF:
@@ -327,36 +365,37 @@ class KafkaConsumerBase(Thread, CypressBase):
                             message.topic(),
                             message.partition(),
                             message.offset()))
+
                     else:
                         # Error
-                        raise confluent_kafka.KafkaError(message.error())
+                        raise Exception(message.error())
                 else:
                     # self.kafka_consumer.commitSync();
                     # self.kafka_consumer.pause();
                     try:
                         message = json.loads(message.value())
-                        response = self.message_handler(message)
-                        if response:
-                            self.message_queue.put((self.producer_topic, response))
+                        if message.get("start_ts"):
+                            self.logger.debug("message spent {}s".format(time() - message["start_ts"]))
+                        self.message_handler(message)
+
                     except:
                         self.logger.error(traceback.format_exc())
             except Exception as ex:
                 self.logger.error(traceback.format_exc())
                 self.logger.error(ex)
-                self.kafka_consumer.close()
-                self.kafka_consumer = None  # reset consumer connection. Try to connect to kafka server again.
+                # try:
+                #     self.kafka_consumer.close()
+                # except:
+                #     self.logger.error(traceback.format_exc())
+                self.reconnect = True  # reset connection. Try to connect to kafka server again.
                 continue
             except KeyboardInterrupt:
                 self.logger.info("Received exit input. Exiting...")
-                try:
-                    if self.kafka_producer:
-                        self.kafka_producer.stop()
-                except:
-                    pass
                 self.running = False
 
         if self.kafka_consumer:
             self.kafka_consumer.close()
+            self.logger.info("close consumer")
 
     def message_handler(self, message):
         """
